"""Sunsync Modbus interface."""
import asyncio
import logging
from typing import Sequence
from urllib.parse import urlparse

import attr
from pymodbus.client.asynchronous.async_io import (  # type: ignore
    AsyncioModbusSerialClient,
    AsyncioModbusTcpClient,
    ModbusClientProtocol,
)

# from pymodbus.client.asynchronous import schedulers  # type: ignore
from pymodbus.client.asynchronous.serial import AsyncModbusSerialClient  # type: ignore
from serial.serialutil import STOPBITS_ONE

from sunsynk.sunsynk import Sunsynk  # type: ignore

_LOGGER = logging.getLogger(__name__)


@attr.define
class pySunsynk(Sunsynk):  # pylint: disable=invalid-name
    """Sunsync Modbus class."""

    port: str = attr.ib(default="/dev/tty0")
    client: ModbusClientProtocol = attr.ib(default=None)

    async def connect(self) -> None:
        """Connect.

        https://pymodbus.readthedocs.io/en/latest/source/example/async_asyncio_serial_client.html

        """
        # pylint: disable=protected-access
        url = urlparse(f"//{self.port}")
        client = None

        if not url.netloc:
            # Cannot run from a coroutine currently
            # https://github.com/riptideio/pymodbus/pull/658/files#r718775308

            # msc = AsyncModbusSerialClient(
            #     schedulers.ASYNC_IO,
            #     port=self.port,
            #     baudrate=self.baudrate,
            #     method="rtu",
            #     stopbits=STOPBITS_ONE,
            #     bytesize=8,
            # )
            # loop, client = msc  # pylint: disable=unpacking-non-sequence

            # Alternative interim...
            client = AsyncioModbusSerialClient(
                port=self.port,
                protocol_class=ModbusClientProtocol,
                framer=AsyncModbusSerialClient._framer(method="rtu"),
                loop=asyncio.get_running_loop(),
                baudrate=self.baudrate,
                stopbits=STOPBITS_ONE,
                bytesize=8,
            )
        else:
            client = AsyncioModbusTcpClient(
                host=url.hostname,
                port=url.port or 502,
                protocol_class=ModbusClientProtocol,
                loop=asyncio.get_running_loop(),
            )

        await client.connect()

        if (hasattr(client, "connected") and not client.connected) or (
            hasattr(client, "_connected") and not client._connected
        ):
            raise ConnectionError

        try:
            client.protocol._timeout = self.timeout
        except AttributeError as err:
            _LOGGER.warning("%s", err)

        self.client = client.protocol

    async def write_register(self, *, address: int, value: int) -> bool:
        """Write to a register - Sunsynk supports modbus function 0x10."""
        try:
            w_r = await self.client.write_registers(
                address=address, values=(value,), unit=self.server_id
            )
            if w_r.function_code < 0x80:  # test that we are not an error
                return True
            _LOGGER.error("failed to write register %s=%s", address, value)
        except asyncio.TimeoutError:
            _LOGGER.error("timeout writing register %s=%s", address, value)
        self.timeouts += 1
        return False

    async def read_holding_registers(self, start: int, length: int) -> Sequence[int]:
        """Read a holding register."""
        res = await self.client.read_holding_registers(
            start, length, unit=self.server_id
        )
        if res.function_code >= 0x80:  # test that we are not an error
            raise Exception(  # pylint: disable=broad-exception-raised
                f"failed to read register {start} - function code: {res.function_code}"
            )
        return res.registers



"""Sensor classes represent modbus registers for an inverter."""
from __future__ import annotations

import logging
from typing import Callable, Generator, Optional, Union

import attr

from sunsynk.helpers import NumType, RegType, SSTime, ValType, as_num
from sunsynk.sensors import Sensor

_LOGGER = logging.getLogger(__name__)
ResolveType = Optional[Callable[[Sensor, ValType], ValType]]


@attr.define(slots=True, eq=False)
class RWSensor(Sensor):
    """Read & write sensor."""

    def reg(self, *regs: int, msg: str = "") -> RegType:
        """Check the registers are within the bitmask."""
        if self.bitmask and regs[0] != (regs[0] & self.bitmask):
            _LOGGER.error(
                "Trying to set a register value outside the sensor's bitmask! sensor:%s regvalue:%s %s",
                self.name,
                regs,
                msg,
            )
            return (regs[0] & self.bitmask,) + tuple(regs[1:])
        return regs

    def value_to_reg(self, value: ValType, resolve: ResolveType) -> RegType:
        """Get the reg value from a display value."""
        raise NotImplementedError()

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the register."""
        return super().reg_to_value(self.masked(regs))

    def __attrs_post_init__(self) -> None:
        """Run post init."""
        if self.bitmask > 0 and len(self.address) != 1:
            _LOGGER.fatal(
                "Sensors with a bitmask should reference a single register! %s [registers=%s]",
                self.name,
                self.address,
            )

    @property
    def dependencies(self) -> list[Sensor]:
        """Dependencies."""
        return []


@attr.define(slots=True, eq=False)
class NumberRWSensor(RWSensor):
    """Numeric sensor which can be read and written."""

    min: int | float | Sensor = attr.field(default=0)
    max: int | float | Sensor = attr.field(default=100)

    @property
    def dependencies(self) -> list[Sensor]:
        """Get a list of sensors upon which this sensor depends."""
        return [s for s in (self.min, self.max) if isinstance(s, Sensor)]

    def value_to_reg(self, value: ValType, resolve: ResolveType) -> RegType:
        """Get the reg value from a display value, or the current reg value if out of range."""
        if value is None or isinstance(value, str):
            raise TypeError
        minv = resolve_num(resolve, self.min, 0)
        maxv = resolve_num(resolve, self.max, 100)
        val = int(max(minv, min(maxv, value / abs(self.factor))))
        if len(self.address) == 1:
            return self.reg(val)
        if len(self.address) == 2:
            return self.reg(val & 0xFFFF, int(val >> 16))
        raise NotImplementedError


@attr.define(slots=True, eq=False)
class SelectRWSensor(RWSensor):
    """Sensor with a set of options to select from."""

    options: dict[int, str] = attr.field(factory=dict)
    switch: Optional[tuple] = attr.field(default=None)

    def __attrs_post_init__(self) -> None:
        """Ensure correct parameters."""
        if self.switch:
            assert self.options == {}
            assert len(self.switch) == 2
            self.options = {self.switch[0]: "OFF", self.switch[1]: "ON"}

    def available_values(self) -> list[str]:
        """Get the available values for this sensor."""
        return list(self.options.values())

    def value_to_reg(self, value: ValType, resolve: ResolveType) -> RegType:
        """Get the reg value from a display value, or the current reg value if out of range."""
        value = str(value)
        regs = [r for r, v in self.options.items() if v == value]
        if regs:
            return self.reg(regs[0])
        _LOGGER.warning("Unknown %s", value)
        current = resolve(self, None) if resolve else 0
        return self.value_to_reg(current, None)  # type:ignore

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the register."""
        regsm = self.masked(regs)
        res = self.options.get(regsm[0])
        if res is None:
            _LOGGER.warning("%s: Unknown register value %s", self.id, regsm[0])
        return res


@attr.define(slots=True, eq=False)
class TimeRWSensor(RWSensor):
    """Extract the time."""

    min: TimeRWSensor = attr.field(default=None)
    max: TimeRWSensor = attr.field(default=None)

    def available_values(
        self, step_minutes: int, resolve: Callable[[Sensor, ValType], ValType]
    ) -> list[str]:
        """Get the available values for this sensor."""
        full_day = 24 * 60

        min_val = SSTime(string=str(resolve(self.min, 0))).minutes if self.min else 0
        max_val = (
            SSTime(string=str(resolve(self.max, 0))).minutes if self.max else full_day
        )
        val = SSTime(string=str(resolve(self, 0))).minutes

        time_range = self._range(min_val, max_val, val, step_minutes, full_day)
        return list(map(lambda m: SSTime(minutes=m).str_value, time_range))

    @property
    def dependencies(self) -> list[Sensor]:
        """Get a list of sensors upon which this sensor depends."""
        return [s for s in (self.min, self.max) if isinstance(s, TimeRWSensor)]

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the time from a register."""
        return SSTime(register=regs[0]).str_value

    def value_to_reg(self, value: ValType, resolve: ResolveType) -> RegType:
        """Get the reg value from a display value."""
        return self.reg(SSTime(string=str(value)).reg_value)

    @staticmethod
    def _range(
        start: int, end: int, val: int, step: int, modulo: int
    ) -> Generator[int, None, None]:
        if val % step != 0:
            yield val
        stop = end if start <= end else end + modulo
        for i in range(start, stop, step):
            yield i % modulo
        if start == end or start != end % modulo:
            yield end


def resolve_num(
    resolve: ResolveType,
    val: Union[NumType, Sensor],
    default: NumType = 0,
) -> NumType:
    """Resolve a number helper."""
    if isinstance(val, (int, float)):
        return val
    if isinstance(val, Sensor):
        res = resolve(val, default) if resolve else 0
        return as_num(res)
    return as_num(val)



"""Test helpers."""
from sunsynk.helpers import SSTime, ensure_tuple, int_round, patch_bitmask, signed
from sunsynk.sensors import Sensor


def test_int_round() -> None:
    res1 = int_round(1.0)
    assert isinstance(res1, int)
    assert res1 == 1


def test_ensure_tuple() -> None:
    assert ensure_tuple(1) == (1,)
    assert ensure_tuple((1,)) == (1,)
    assert ensure_tuple((1, 5)) == (1, 5)
    assert ensure_tuple("a") == ("a",)


def test_signed() -> None:
    assert signed(0x7FFF) == 0x7FFF
    assert signed(0xFFFF) == 0


def test_signeds() -> None:
    """Signed sensors have a -1 factor"""
    s = Sensor(1, "", "", factor=-1)
    assert s.reg_to_value((1,)) == 1
    assert s.reg_to_value((0xFFFE,)) == -1

    s = Sensor(1, "", "", factor=1)
    assert s.reg_to_value((1,)) == 1
    assert s.reg_to_value((0xFFFE,)) == 0xFFFE
    assert s.reg_to_value((1, 1)) == 0x10001


def test_time() -> None:
    time = SSTime(minutes=10)
    assert time.str_value == "0:10"
    assert time.reg_value == 10
    time.str_value = "0:10"
    assert time.minutes == 10
    time.str_value = "00:10"
    assert time.minutes == 10
    time.reg_value = 10
    assert time.minutes == 10

    time = SSTime(minutes=100)
    assert time.str_value == "1:40"
    assert time.reg_value == 140
    time.str_value = "1:40"
    assert time.minutes == 100
    time.str_value = "01:40"
    assert time.minutes == 100
    time.reg_value = 140
    assert time.minutes == 100

    just_before_midnight = 23 * 60 + 59
    time = SSTime(minutes=just_before_midnight)
    assert time.str_value == "23:59"
    assert time.reg_value == 2359
    time.str_value = "23:59"
    assert time.minutes == just_before_midnight
    time.reg_value = 2359
    assert time.minutes == just_before_midnight


def test_patch_bitmask() -> None:
    assert patch_bitmask(2, 1, 1) == 3
    assert patch_bitmask(1, 2, 2) == 3

    assert patch_bitmask(0xFFF, 0, 1) == 0xFFE
    assert patch_bitmask(0xFFFF, 0, 1) == 0xFFFE
    assert patch_bitmask(0xFFF, 0, 2) == 0xFFD



"""Optionally test filters."""
import logging
from itertools import repeat
from pathlib import Path
from typing import Any, Callable, List

import pytest

from sunsynk.definitions import SENSORS
from sunsynk.rwsensors import RWSensor
from tests.conftest import import_module

_LOGGER = logging.getLogger(__name__)


@pytest.fixture
def mod_folder() -> str:
    """Return the folder from __init__."""
    init = import_module("__init__", Path(__file__).parent)
    return init.MOD_FOLDER


@pytest.fixture
def filters(mod_folder) -> Callable:
    """Import & return the getfilter function."""
    return import_module("filter", mod_folder)


def assert_sequence(fut, updates: List[Any], counts=None):
    upd_cnt, i_cnt = 0, 0
    try:
        for upd in updates:
            i_cnt += 1
            if upd is None:
                assert not fut.should_update()
                continue

            assert fut.should_update()
            upd_cnt += 1

            if isinstance(upd, tuple):  # expecting a result!
                ures = fut.update(upd[0])
                if ures is None or ures != upd[1]:
                    assert 0, f"{upd}: update({upd[0]}) expected {upd[1]} got {ures}"
            else:
                ures = fut.update(upd)
                if ures is not None:
                    assert 0, f"update({upd}) expected None, got {ures}"

    except AssertionError as err:
        assert False, f"Index {i_cnt} Update {upd_cnt} - {err}"

    if counts:
        assert counts == (i_cnt, upd_cnt)


def add_intervals(count, val0, *vals):
    """Generator to insert count intervals between values of an array."""
    assert not isinstance(val0, list)
    yield val0
    for val in vals:
        # an interval is represented as None
        yield from repeat(None, count)
        yield val


def test_last(filters):
    """Last filter."""
    fut = filters.getfilter("last", None)

    assert_sequence(fut, add_intervals(59, *[(55, 55), (44, 44), (43, 43)]))


def test_min(filters):
    """Min filter."""
    fut = filters.getfilter("min", None)

    sq0 = (50, 50)
    sq1 = [44, 100, 100, 100, 100, (100, 44)]
    sq2 = [100, 100, 100, 55, 100, (100, 55)]

    assert_sequence(fut, add_intervals(9, sq0, sq1, sq2))

    assert fut.should_update() is False


def test_mean(filters):
    """Mean filter."""
    fut = filters.getfilter("mean", None)

    assert_sequence(
        fut,
        add_intervals(9, *[(50, 50), 50, 50, 50, 100, 100, (100, 75)]),
        counts=(61, 7),
    )

    assert_sequence(  # a single interval
        fut,
        [None] * 9,
        counts=(9, 0),
    )

    assert_sequence(
        fut,
        add_intervals(9, *[50, 100, 50, 50, 100, (100, 75)]),
        counts=(51, 6),
    )


def test_step(filters):
    """Step filter."""
    fut = filters.getfilter("", None)
    assert_sequence(
        fut,
        [(20, 20), 20, 20, 20, (120, 120), 140, 140],
    )

    fut = filters.getfilter("step:100", None)
    assert_sequence(
        fut,
        [(90, 90)] + [90] * 58 + [(90, 90)],
    )

    fut = filters.getfilter("step:100", None)
    sq0 = [(10, 10)]
    sq0 += [21] * 58 + [(50, 21.3)]
    sq0 += [(150, 150), (20, 20)]
    sq0 += [21] * 58 + [(50, 21.5)]
    sq0 += [20] * 58 + [(50, 20.5)]
    sq0 += [20] * 58 + [(50, 20.5)]
    sq0 += [(1000, 1000)] + [950] * 58 + [(950, 950.8)]

    assert_sequence(fut, sq0)


def test_step_text(filters):
    """Step filter."""
    fut = filters.getfilter("step", None)
    assert_sequence(
        fut,
        [("a", "a"), ("b", "b")],
    )


def test_rr(filters):
    """Test round robin."""
    RROBIN = filters.RROBIN
    RROBIN.tick()

    sen = [filters.Sensor(1, "a"), filters.Sensor(2, "b"), filters.Sensor(3, "c")]
    fil = [filters.getfilter("round_robin", s) for s in sen]

    for idx in range(2):
        assert fil[idx].sensor_name is sen[idx].name
    assert RROBIN.list == list(fil)

    assert [RROBIN.idx, RROBIN.active] == [-1, []]
    RROBIN.tick()
    assert [RROBIN.idx, RROBIN.active] == [0, [fil[0]]]

    # Cycle through should_update()
    assert [f.should_update() for f in fil] == [True, False, False]
    RROBIN.tick()
    assert [f.should_update() for f in fil] == [False, True, False]
    RROBIN.tick()
    assert [f.should_update() for f in fil] == [False, False, True]
    RROBIN.tick()
    assert [f.should_update() for f in fil] == [True, False, False]


def test_suggest(filters):
    assert filters.suggested_filter(SENSORS.all["environment_temperature"]) == "avg"
    assert filters.suggested_filter(SENSORS.all["day_battery_charge"]) == "last"
    # assert filters.suggested_filter(SENSORS.all["grid_ct_load"]) == "step"
    assert filters.suggested_filter(SENSORS.all["sd_status"]) == "last"

    rw_sensors = [s for s in SENSORS.all.values() if isinstance(s, RWSensor)]
    for s in rw_sensors:
        assert filters.suggested_filter(s) == "round_robin"



"""Test sunsynk library."""
import logging
import os
from unittest.mock import AsyncMock, MagicMock, call, patch

import pytest

# from sunsynk.definitions import serial
from sunsynk import Sunsynk
from sunsynk.pysunsynk import pySunsynk
from sunsynk.rwsensors import NumberRWSensor
from sunsynk.state import InverterState
from sunsynk.usunsynk import uSunsynk

_LOGGER = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_pyss():
    ss = pySunsynk()
    with pytest.raises(ConnectionError):
        await ss.connect()


@pytest.mark.asyncio
async def test_uss_schemes():
    """Test url schemes for usunsynk.

    umodbus only connects on read.
    """
    for port in ("serial:///dev/usb1", "tcp://127.0.0.1:502"):
        ss = uSunsynk(port=port)
        try:
            await ss.connect()
        except ModuleNotFoundError as err:  # not working on windows
            _LOGGER.error("usunsynk could not connect to %s: %s", port, err)
            if os.name == "posix":
                raise

    for port in ("127.0.0.1:502", "xxx", "localhost"):
        ss = uSunsynk(port=port)
        with pytest.raises(ValueError):
            await ss.connect()


@pytest.mark.asyncio
async def test_uss_sensor():
    ss = uSunsynk(port="tcp://127.0.0.1:502")
    await ss.connect()

    rhr = ss.client.read_holding_registers = AsyncMock()

    _LOGGER.warning("%s", dir(ss.client))
    assert not rhr.called
    await ss.read_holding_registers(1, 2)
    assert rhr.called

    wrr = ss.client.write_registers = AsyncMock()
    assert not wrr.called
    await ss.write_register(address=1, value=2)
    assert wrr.called


@pytest.mark.asyncio
async def test_ss_NotImplemented():
    ss = Sunsynk()
    with pytest.raises(NotImplementedError):
        await ss.connect()
    with pytest.raises(NotImplementedError):
        await ss.write_register(address=1, value=1)
    with pytest.raises(NotImplementedError):
        await ss.read_holding_registers(1, 1)


@pytest.mark.asyncio
@patch("sunsynk.Sunsynk.read_holding_registers")
@patch("sunsynk.Sunsynk.write_register")
async def test_ss_write_sensor(
    wreg: MagicMock, rhr: MagicMock, state: InverterState, caplog
):
    ss = Sunsynk()
    ss.state = state

    sen = NumberRWSensor((1,), "s1")
    state.track(sen)

    assert sen.value_to_reg(44, state.get) == (44,)

    await ss.write_sensor(sen, 44)
    assert state.registers == {1: 44}

    wreg.assert_called_once()
    assert wreg.call_args == call(address=1, value=44)

    # test a sensor with a bitmask
    sen = NumberRWSensor((1,), "s2", min=1, max=10, bitmask=0x3)
    state.track(sen)
    rhr.return_value = (1,)

    await ss.write_sensor(sen, 3)
    assert state.registers == {1: 3}

    await ss.write_sensor(sen, 5)
    assert state.registers == {1: 1}
    assert "outside" in caplog.text


@pytest.mark.asyncio
@patch("sunsynk.Sunsynk.read_holding_registers")
@patch("sunsynk.Sunsynk.write_register")
async def test_ss_write_sensor_bm(
    wreg: MagicMock, rhr: MagicMock, state: InverterState, caplog
):
    ss = Sunsynk()
    ss.state = state

    sen = NumberRWSensor((1,), "s1", bitmask=0x10)
    state.track(sen)
    state.update({1: 0xFF})
    assert state.registers == {1: 0xFF}

    assert sen.value_to_reg(0x10, state.get) == (0x10,)
    assert "outside" not in caplog.text

    rhr.return_value = [0xF8]  # write will perform a read!
    await ss.write_sensor(sen, 0x10)
    assert "outside" not in caplog.text
    assert state.registers == {1: 0xF8}
    wreg.assert_called_once()
    assert wreg.call_args == call(address=1, value=0xF8)

    await ss.write_sensor(sen, 0x00)
    assert state.registers == {1: 0xE8}
    assert "outside" not in caplog.text


@pytest.mark.asyncio
@patch("sunsynk.Sunsynk.read_holding_registers")
async def test_ss_read_sensors(rhr: MagicMock, state: InverterState):
    ss = Sunsynk()
    ss.state = state
    sen = NumberRWSensor((1,), "", min=1, max=10)
    ss.state.track(sen)

    rhr.return_value = (1,)
    assert ss.state[sen] is None
    await ss.read_sensors([sen])
    assert ss.state[sen] == 1

    rhr.side_effect = Exception("a")
    with pytest.raises(Exception):
        await ss.read_sensors([sen])



"""Sunsync Modbus interface."""
import asyncio
import logging
from typing import Iterable, Sequence

import attr

from sunsynk.helpers import patch_bitmask
from sunsynk.rwsensors import RWSensor
from sunsynk.sensors import Sensor, ValType
from sunsynk.state import InverterState, group_sensors, register_map

_LOGGER = logging.getLogger(__name__)


@attr.define
class Sunsynk:
    """Sunsync Modbus class."""

    state: InverterState = attr.field(factory=InverterState)
    port: str = attr.ib(default="/dev/tty0")
    baudrate: int = attr.ib(default=9600)
    server_id: int = attr.ib(default=1)
    timeout: int = attr.ib(default=10)
    read_sensors_batch_size: int = attr.field(default=60)
    timeouts: int = 0

    async def connect(self) -> None:
        """Connect."""
        raise NotImplementedError

    async def write_register(self, *, address: int, value: int) -> bool:
        """Write to a register - Sunsynk support function code 0x10."""
        raise NotImplementedError

    async def write_sensor(
        self, sensor: RWSensor, value: ValType, *, msg: str = ""
    ) -> None:
        """Write a sensor."""
        regs = sensor.value_to_reg(value, self.state.get)
        # if bitmask we should READ the register first!!!
        if sensor.bitmask:
            _LOGGER.debug("0 - %s", regs)
            regs = sensor.reg(*regs, msg=f"while setting value = {value}")
            _LOGGER.debug("1 - %s", regs)
            val1 = regs[0]
            r_r = await self.read_holding_registers(sensor.address[0], 1)
            _LOGGER.debug("r_r - %s", r_r)
            val0 = r_r[0]
            regs0 = patch_bitmask(val0, val1, sensor.bitmask)
            regs = (regs0,) + regs[1:]
            msg = f"[Register {val0}-->{val1}]"

        _LOGGER.critical(
            "Writing sensor %s=%s [%s=%s] %s",
            sensor.id,
            value,
            sensor.address,
            regs,
            msg,
        )
        for idx, addr in enumerate(sensor.address):
            if idx:
                await asyncio.sleep(0.05)
            await self.write_register(address=addr, value=regs[idx])
            self.state.registers[addr] = regs[idx]

    async def read_holding_registers(self, start: int, length: int) -> Sequence[int]:
        """Read a holding register."""
        raise NotImplementedError

    async def read_sensors(self, sensors: Iterable[Sensor]) -> None:
        """Read a list of sensors - Sunsynk supports function code 0x03."""
        # Check if state is ok & tracking the sensors being read
        assert self.state is not None
        for sen in sensors:
            if sen not in self.state.values:
                _LOGGER.warning("sensor %s not being tracked", sen.id)

        new_regs: dict[int, int] = {}
        for grp in group_sensors(
            sensors, allow_gap=1, max_group_size=self.read_sensors_batch_size
        ):
            glen = grp[-1] - grp[0] + 1
            try:
                r_r = await asyncio.wait_for(
                    self.read_holding_registers(grp[0], glen), timeout=self.timeout + 1
                )
            except asyncio.TimeoutError:
                _LOGGER.error("timeout reading register %s (%s)", grp[0], glen)
                self.timeouts += 1
                raise
            except Exception as err:  # pylint: disable=broad-except
                raise Exception(  # pylint: disable=raise-missing-from,broad-exception-raised
                    f"({self.server_id},{grp[0]},{glen}) {err}"
                )

            regs = register_map(grp[0], r_r)
            new_regs.update(regs)

            if len(r_r) != glen:
                _LOGGER.warning(
                    "Did not complete read, only read %s/%s", len(r_r), glen
                )

            _LOGGER.debug(
                "Request registers: %s glen=%d. Response %s len=%d. regs=%s",
                grp,
                glen,
                r_r,
                len(r_r),
                regs,
            )

        self.state.update(new_regs)



"""Sunsynk library."""
from typing import Final

# flake8: noqa
from .sensors import Sensor  # pylint: disable=unused-import
from .sunsynk import Sunsynk  # pylint: disable=unused-import

VERSION = "0.3.1"

CELSIUS: Final = "°C"
KWH: Final = "kWh"
AMPS: Final = "A"
VOLT: Final = "V"
WATT: Final = "W"



"""Helper functions."""
import logging
from math import modf
from typing import Any, Optional, Tuple, Union

_LOGGER = logging.getLogger(__name__)

ValType = Union[float, int, str, None]
RegType = tuple[int, ...]
NumType = Union[float, int]


def ensure_tuple(val: Any) -> Tuple[int]:
    """Return a tuple."""
    if isinstance(val, tuple):
        return val  # type: ignore
    if isinstance(val, int):
        return (val,)
    return tuple(val)  # type: ignore


def int_round(val: NumType) -> NumType:
    """Round if float."""
    if not isinstance(val, float):
        return val
    val = round(val, 2)
    if modf(val)[0] == 0:
        return int(val)
    return val


def as_num(val: ValType) -> Union[float, int]:
    """Convert to float."""
    if isinstance(val, (float, int)):
        return val
    if val is None:
        return 0
    try:
        return int(val)
    except ValueError:
        pass
    try:
        return float(val)
    except ValueError as err:
        _LOGGER.error(str(err))
    return 0


def signed(val: Union[int, float]) -> Union[int, float]:
    """Convert 16-bit value to signed int."""
    return val if val <= 0x7FFF else val - 0xFFFF


def slug(name: str) -> str:
    """Create a slug."""
    return name.lower().replace(" ", "_").replace("-", "_")


class SSTime:
    """Deals with inverter time format conversion complexities."""

    minutes: int = 0

    def __init__(
        self,
        *,
        minutes: Optional[int] = None,
        register: Optional[int] = None,
        string: Optional[str] = None,
    ) -> None:
        """Init the time. All mutually exclusive."""
        if minutes is not None:
            assert register is None
            assert string is None
            self.minutes = minutes
        elif register is not None:
            assert string is None
            self.reg_value = register
        elif string is not None:
            self.str_value = string

    @property
    def reg_value(self) -> int:
        """Get the register value."""
        hours, minutes = divmod(self.minutes, 60)
        return hours * 100 + minutes

    @reg_value.setter
    def reg_value(self, reg_value: int) -> None:
        """Convert from a register value."""
        hours, minutes = divmod(reg_value, 100)
        self.minutes = hours * 60 + minutes

    @property
    def str_value(self) -> str:
        """Get the value in hh:mm format."""
        hours, minutes = divmod(self.minutes, 60)
        return f"{hours}:{minutes:02}"

    @str_value.setter
    def str_value(self, value: str) -> None:
        """Parse a string in hh:mm format."""
        (hours, _, minutes) = value.partition(":")
        self.minutes = int(hours) * 60 + int(minutes)


def patch_bitmask(value: int, patch: int, bitmask: int) -> int:
    """Combine bitmask values."""
    return (patch & bitmask) + (value & (0xFFFF - bitmask))



"""Test run."""
import logging
import re
from pathlib import Path
from types import ModuleType

import pytest

from tests.conftest import import_module

_LOGGER = logging.getLogger(__name__)


@pytest.fixture
def mod_folder() -> str:
    """Return the folder from __init__."""
    init = import_module("__init__", Path(__file__).parent)
    return init.MOD_FOLDER


@pytest.fixture
def run(mod_folder) -> ModuleType:
    """Import the module."""
    return import_module("run", mod_folder)


def test_run(run):
    """Test Run."""
    assert not run.STATES
    assert not run.OPT.mqtt_host

    # testargs = ["run.py", "host1", "passw"]
    # with patch.object(run.sys, "argv", testargs):
    #     run.startup()
    # assert run.STATES
    # assert run.OPT.mqtt_host == "host1"
    # assert run.OPT.mqtt_password == "passw"

    # run.STATES.clear()
    # run.OPT.mqtt_host = ""


def test_versions(run, mod_folder):
    """Test versions.

    config.json - contains the HASS addon version
    Dockerfile  - installs the specific sunsynk library from pypi
    setup.py    - sunsynk library on pypi
    """

    def _get_version(filename, regex):
        txt = Path(filename).read_text()
        res = re.compile(regex).search(txt)
        assert res, f"version not found in {filename}"
        return res.group(1)

    v_setup = _get_version(
        filename="sunsynk/__init__.py",
        regex=r'VERSION = "(.+)"',
    )

    v_docker = _get_version(
        filename=f"{mod_folder}/Dockerfile",
        regex=r"sunsynk(?:\[[^\]]+\])?==([0-9.]+)",
    )

    v_config = _get_version(
        filename=f"{mod_folder}/config.yaml",
        regex=r"version: .+-(.+)",
    )

    assert v_setup == v_docker
    assert v_setup == v_config



"""Register state."""
import logging

# from collections.abc import KeysView
from typing import Callable, Generator, Iterable, Iterator, Optional, Sequence

import attr

from sunsynk.rwsensors import RWSensor
from sunsynk.sensors import Sensor, ValType

_LOGGER = logging.getLogger(__name__)


@attr.define()
class InverterState:
    """Keep the state of the inverter."""

    values: dict[Sensor, ValType] = attr.field(factory=dict)
    registers: dict[int, int] = attr.field(factory=dict)
    onchange: Optional[Callable[[Sensor, ValType, ValType], None]] = attr.field(
        default=None
    )

    def __getitem__(self, sensor: Sensor) -> ValType:
        """Get the current value of a sensor."""
        return self.values.get(sensor)

    def get(self, sensor: Sensor, default: ValType = None) -> ValType:
        """Get the current value of a sensor."""
        return self.values.get(sensor, default)

    def track(self, *sensor: Sensor) -> None:
        """Add a sensor to be tracked."""
        for sen in sensor:
            self.values.setdefault(sen, None)
            if isinstance(sen, RWSensor) and sen.dependencies:
                for dep in sen.dependencies:
                    self.values.setdefault(dep, None)

    @property
    def sensors(self) -> Iterator[Sensor]:
        """Get a generator of all sensors."""
        return iter(self.values.keys())

    def update(self, new_regs: dict[int, int]) -> None:
        """Update the state."""
        changed: dict[Sensor, tuple[ValType, ValType]] = {}  # sensor, old & new value

        for sen in self.sensors:
            # No updates for this sensor, skip
            if not any(a in new_regs for a in sen.address):
                continue

            regs = tuple(new_regs.get(a, self.registers.get(a, 0)) for a in sen.address)

            assert isinstance(regs, tuple)
            assert isinstance(sen.address, tuple)
            assert len(regs) == len(sen.address)

            oldv = self.values[sen]
            if sen.bitmask:
                regs = (regs[0] & sen.bitmask,)

            newv = sen.reg_to_value(regs)
            _LOGGER.debug("register %s = %s (old=%s)", self.registers, oldv, newv)
            if oldv != newv:
                self.values[sen] = newv
                changed[sen] = (newv, oldv)

        self.registers.update(new_regs)

        if not self.onchange:
            return
        for sen, (new, old) in changed.items():
            self.onchange(sen, new, old)


def group_sensors(
    sensors: Iterable[Sensor], allow_gap: int = 3, max_group_size: int = 60
) -> Generator[list[int], None, None]:
    """Group sensor registers into blocks for reading."""
    if not sensors:
        return
    regs = {r for s in sensors for r in s.address}
    group: list[int] = []
    adr0 = 0
    for adr1 in sorted(regs):
        if group and (adr1 - adr0 > allow_gap or len(group) >= max_group_size):
            yield group
            group = []
        adr0 = adr1
        group.append(adr1)
    if group:
        yield group


def register_map(start: int, registers: Sequence[int]) -> dict[int, int]:
    """Turn the registers into a dictionary or map."""
    return dict(enumerate(registers, start))



"""Sensor classes represent modbus registers for an inverter."""
from __future__ import annotations

import logging
from typing import Union

import attr

from sunsynk.helpers import (
    NumType,
    RegType,
    ValType,
    ensure_tuple,
    int_round,
    signed,
    slug,
)

_LOGGER = logging.getLogger(__name__)


@attr.define(slots=True)
class Sensor:
    """Sunsynk sensor."""

    # pylint: disable=too-many-instance-attributes
    address: RegType = attr.field(converter=ensure_tuple)
    name: str = attr.field()
    unit: str = attr.field(default="")
    factor: float = attr.field(default=1)
    bitmask: int = attr.field(default=0)

    @property
    def id(self) -> str:  # pylint: disable=invalid-name
        """Get the sensor ID."""
        return slug(self.name)

    def reg_to_value(self, regs: RegType) -> ValType:
        """Return the value from the registers."""
        regs = self.masked(regs)
        val: NumType = regs[0]
        if len(regs) == 2:
            val += regs[1] << 16
        elif self.factor < 0:  # Indicates this register is signed
            val = signed(val)
        val = int_round(val * abs(self.factor))
        _LOGGER.debug("%s=%s%s %s", self.id, val, self.unit, regs)
        return val

    def masked(self, regs: RegType) -> RegType:
        """Return the masked reg."""
        if self.bitmask:
            return tuple(r & self.bitmask for r in regs)
        return regs

    def __hash__(self) -> int:
        """Hash the sensor id."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Sensor equality is based on the ID only."""
        if not isinstance(other, Sensor):
            raise TypeError
        return self.id == other.id


@attr.define(slots=True)
class SensorDefinitions:
    """Definitions."""

    all: dict[str, Sensor] = attr.field(factory=dict)
    deprecated: dict[str, Sensor] = attr.field(factory=dict)

    @property
    def serial(self) -> Sensor:
        """Get the serial sensor."""
        return self.all["serial"]

    @property
    def rated_power(self) -> Sensor:
        """Get the rated power sensor."""
        return self.all["rated_power"]

    def __add__(
        self, item: Union[Sensor, tuple[Sensor, ...], list[Sensor]]
    ) -> SensorDefinitions:
        """Add new sensors."""
        if isinstance(item, Sensor):
            self.all[item.id] = item
            return self
        if isinstance(item, (tuple, list)):
            for itm in item:
                self.all[itm.id] = itm
        return self


@attr.define(slots=True, eq=False)
class MathSensor(Sensor):
    """Math sensor, add multiple registers."""

    factors: tuple[float, ...] = attr.field(default=None, converter=ensure_tuple)
    no_negative: bool = attr.field(default=False)
    absolute: bool = attr.field(default=False)

    def reg_to_value(self, regs: RegType) -> ValType:
        """Calculate the math value."""
        val = int_round(sum(signed(i) * s for i, s in zip(regs, self.factors)))
        if self.absolute and val < 0:
            val = -val
        if self.no_negative and val < 0:
            val = 0
        return val

    def __attrs_post_init__(self) -> None:
        """Ensure correct parameters."""
        assert len(self.address) == len(self.factors)


@attr.define(slots=True, eq=False)
class TempSensor(Sensor):
    """Offset by 100 for temperature."""

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the temperature (offset by 100)."""
        try:
            val = regs[0]
            return int_round((float(val) * abs(self.factor)) - 100)  # type: ignore
        except (TypeError, ValueError) as err:
            _LOGGER.error("Could not decode temperature: %s", err)
        return None


@attr.define(slots=True, eq=False)
class SDStatusSensor(Sensor):
    """SD card status."""

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the SD card status."""
        return {
            1000: "fault",
            2000: "ok",
        }.get(regs[0]) or f"unknown {regs[0]}"


@attr.define(slots=True, eq=False)
class InverterStateSensor(Sensor):
    """Inverter status."""

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the inverter status."""
        if regs[0] == 2:
            return "ok"
        return f"unknown {regs[0]}"


@attr.define(slots=True, eq=False)
class SerialSensor(Sensor):
    """Decode the inverter serial number."""

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode the inverter serial number."""
        val = ""
        for b16 in regs:
            val += chr(b16 >> 8)
            val += chr(b16 & 0xFF)
        return val


@attr.define(slots=True, eq=False)
class FaultSensor(Sensor):
    """Decode Inverter faults."""

    def reg_to_value(self, regs: RegType) -> ValType:
        """Decode Inverter faults."""
        faults = {
            13: "Working mode change",
            18: "AC over current",
            20: "DC over current",
            23: "F23 AC leak current or transient over current",
            24: "F24 DC insulation impedance",
            26: "F26 DC busbar imbalanced",
            29: "Parallel comms cable",
            35: "No AC grid",
            42: "AC line low voltage",
            47: "AC freq high/low",
            56: "DC busbar voltage low",
            63: "ARC fault",
            64: "Heat sink tempfailure",
        }
        err = []
        off = 0
        for b16 in regs:
            for bit in range(16):
                msk = 1 << bit
                if msk & b16:
                    msg = f"F{bit+off+1:02} " + faults.get(off + msk, "")
                    err.append(msg.strip())
            off += 16
        return ", ".join(err)



"""Global fixtures."""
from unittest.mock import MagicMock

import pytest

from sunsynk.state import InverterState


@pytest.fixture
def state() -> InverterState:
    return InverterState(onchange=MagicMock())



"""Sunsynk 5kW&8kW hybrid inverter sensor definitions."""
from sunsynk import AMPS, CELSIUS, KWH, VOLT, WATT
from sunsynk.rwsensors import NumberRWSensor, SelectRWSensor, TimeRWSensor
from sunsynk.sensors import (
    FaultSensor,
    InverterStateSensor,
    MathSensor,
    SDStatusSensor,
    Sensor,
    SensorDefinitions,
    SerialSensor,
    TempSensor,
)

SENSORS = SensorDefinitions()

##########
# Battery
##########
SENSORS += (
    TempSensor(182, "Battery temperature", CELSIUS, 0.1),
    Sensor(183, "Battery voltage", VOLT, 0.01),
    Sensor(184, "Battery SOC", "%"),
    Sensor(190, "Battery power", WATT, -1),
    Sensor(191, "Battery current", AMPS, -0.01),
)

#################
# Inverter Power
#################
SENSORS += (
    Sensor(175, "Inverter power", WATT, -1),
    Sensor(154, "Inverter voltage", VOLT, 0.1),
    Sensor(193, "Inverter frequency", "Hz", 0.01),
)

#############
# Grid Power
#############
SENSORS += (
    Sensor(79, "Grid frequency", "Hz", 0.01),
    Sensor(169, "Grid power", WATT, -1),  # L1(167) + L2(168)
    Sensor(167, "Grid LD power", WATT, -1),  # L1 seems to be LD
    Sensor(168, "Grid L2 power", WATT, -1),
    Sensor(150, "Grid voltage", VOLT, 0.1),
    MathSensor((160, 161), "Grid current", AMPS, factors=(0.01, 0.01)),
    Sensor(172, "Grid CT power", WATT, -1),
)
# LD power?

#############
# Load Power
#############
SENSORS += (
    Sensor(178, "Load power", WATT, -1),  # L1(176) + L2(177)
    Sensor(176, "Load L1 power", WATT, -1),
    Sensor(177, "Load L2 power", WATT, -1),
)

################
# Solar Power 1
################
SENSORS += (
    Sensor(186, "PV1 power", WATT, -1),
    Sensor(109, "PV1 voltage", VOLT, 0.1),
    Sensor(110, "PV1 current", AMPS, 0.1),
)

################
# Solar Power 2
################
SENSORS += (
    Sensor(187, "PV2 power", WATT, -1),
    Sensor(111, "PV2 voltage", VOLT, 0.1),
    Sensor(112, "PV2 current", AMPS, 0.1),
)

###################
# Power on Outputs
###################
SENSORS += (
    Sensor(166, "AUX power", WATT, -1),
    # https://github.com/kellerza/sunsynk/issues/75
    #  https://powerforum.co.za/topic/8646-my-sunsynk-8kw-data-collection-setup/?do=findComment&comment=147591
    MathSensor(
        (175, 169, 166), "Essential power", WATT, factors=(1, 1, -1), absolute=True
    ),
    # MathSensor((175, 167, 166), "Essential power", WATT, factors=(1, 1, -1)),
    MathSensor(
        (172, 167), "Non-Essential power", WATT, factors=(1, -1), no_negative=True
    ),
)

###################
# Energy
###################
SENSORS += (
    Sensor(60, "Day Active Energy", KWH, -0.1),
    Sensor(70, "Day Battery Charge", KWH, 0.1),
    Sensor(71, "Day Battery discharge", KWH, 0.1),
    Sensor(77, "Day Grid Export", KWH, 0.1),
    Sensor(76, "Day Grid Import", KWH, 0.1),
    # Sensor(200, "Day Load Power", KWH, 0.01),
    Sensor(84, "Day Load Energy", KWH, 0.1),
    Sensor(108, "Day PV Energy", KWH, 0.1),
    Sensor(61, "Day Reactive Energy", "kVarh", -0.1),
    # Sensor((201, 202), "History Load Power", KWH, 0.1),
    Sensor(67, "Month Grid Energy", KWH, 0.1),
    Sensor(66, "Month Load Energy", KWH, 0.1),
    Sensor(65, "Month PV Energy", KWH, 0.1),
    Sensor((63, 64), "Total Active Energy", KWH, 0.1),  # signed?
    Sensor((72, 73), "Total Battery Charge", KWH, 0.1),
    Sensor((74, 75), "Total Battery Discharge", KWH, 0.1),
    Sensor((81, 82), "Total Grid Export", KWH, 0.1),
    Sensor((78, 80), "Total Grid Import", KWH, 0.1),
    Sensor((85, 86), "Total Load Energy", KWH, 0.1),
    Sensor((96, 97), "Total PV Energy", KWH, 0.1),
    Sensor((98, 99), "Year Grid Export", KWH, 0.1),
    Sensor((87, 88), "Year Load Energy", KWH, 0.1),
    Sensor((68, 69), "Year PV Energy", KWH, 0.1),
)

##########
# General
##########
RATED_POWER = Sensor((16, 17), "Rated power", WATT, 0.1)
SENSORS += (
    RATED_POWER,
    SerialSensor((3, 4, 5, 6, 7), "Serial"),
    Sensor(0, "Device Type"),
    FaultSensor((103, 104, 105, 106), "Fault"),
    InverterStateSensor(59, "Overall state"),
    SDStatusSensor(92, "SD Status", ""),  # type: ignore
    TempSensor(90, "DC transformer temperature", CELSIUS, 0.1),
    TempSensor(95, "Environment temperature", CELSIUS, 0.1),
    TempSensor(91, "Radiator temperature", CELSIUS, 0.1),
    Sensor(194, "Grid Connected Status"),
)
###########
# Settings
###########
SENSORS += (
    Sensor(200, "Control Mode"),
    Sensor(232, "Grid Charge enabled", "", -1),
    Sensor(312, "Battery charging voltage", VOLT, 0.01),
    Sensor(603, "Bat1 SOC", "%"),
    Sensor(611, "Bat1 Cycle"),
)

# Absolute min and max voltage based on Deye inverter
MIN_VOLTAGE = 41
MAX_VOLTAGE = 60

BATTERY_EQUALIZATION_VOLTAGE = NumberRWSensor(
    201, "Battery Equalization voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)
BATTERY_ABSORPTION_VOLTAGE = NumberRWSensor(
    202, "Battery Absorption voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)
BATTERY_FLOAT_VOLTAGE = NumberRWSensor(
    203, "Battery Float voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)

BATTERY_SHUTDOWN_CAPACITY = NumberRWSensor(217, "Battery Shutdown Capacity", "%")
BATTERY_RESTART_CAPACITY = NumberRWSensor(218, "Battery Restart Capacity", "%")
BATTERY_LOW_CAPACITY = NumberRWSensor(
    219,
    "Battery Low Capacity",
    "%",
    min=BATTERY_SHUTDOWN_CAPACITY,
    max=BATTERY_RESTART_CAPACITY,
)
BATTERY_SHUTDOWN_CAPACITY.max = BATTERY_LOW_CAPACITY
BATTERY_RESTART_CAPACITY.min = BATTERY_LOW_CAPACITY

BATTERY_SHUTDOWN_VOLTAGE = NumberRWSensor(
    220, "Battery Shutdown voltage", VOLT, 0.01, min=MIN_VOLTAGE
)
BATTERY_RESTART_VOLTAGE = NumberRWSensor(
    221, "Battery Restart voltage", VOLT, 0.01, max=MAX_VOLTAGE
)
BATTERY_LOW_VOLTAGE = NumberRWSensor(
    222,
    "Battery Low voltage",
    VOLT,
    0.01,
    min=BATTERY_SHUTDOWN_VOLTAGE,
    max=BATTERY_RESTART_VOLTAGE,
)
BATTERY_SHUTDOWN_VOLTAGE.max = BATTERY_LOW_VOLTAGE
BATTERY_RESTART_VOLTAGE.min = BATTERY_LOW_VOLTAGE

SENSORS += (
    BATTERY_EQUALIZATION_VOLTAGE,
    BATTERY_ABSORPTION_VOLTAGE,
    BATTERY_FLOAT_VOLTAGE,
    BATTERY_SHUTDOWN_CAPACITY,
    BATTERY_RESTART_CAPACITY,
    BATTERY_LOW_CAPACITY,
    BATTERY_SHUTDOWN_VOLTAGE,
    BATTERY_RESTART_VOLTAGE,
    BATTERY_LOW_VOLTAGE,
)

#################
# System program
#################
SENSORS += SelectRWSensor(
    243, "Priority Mode", options={0: "Battery first", 1: "Load first"}
)

SENSORS += SelectRWSensor(
    244,
    "Load Limit",
    options={0: "Allow Export", 1: "Essentials", 2: "Zero Export"},
)

PROG1_TIME = TimeRWSensor(250, "Prog1 Time")
PROG2_TIME = TimeRWSensor(251, "Prog2 Time", min=PROG1_TIME)
PROG3_TIME = TimeRWSensor(252, "Prog3 Time", min=PROG2_TIME)
PROG4_TIME = TimeRWSensor(253, "Prog4 Time", min=PROG3_TIME)
PROG5_TIME = TimeRWSensor(254, "Prog5 Time", min=PROG4_TIME)
PROG6_TIME = TimeRWSensor(255, "Prog6 Time", min=PROG5_TIME)
PROG1_TIME.min = PROG6_TIME
PROG1_TIME.max = PROG2_TIME
PROG2_TIME.max = PROG3_TIME
PROG3_TIME.max = PROG4_TIME
PROG4_TIME.max = PROG5_TIME
PROG5_TIME.max = PROG6_TIME
PROG6_TIME.max = PROG1_TIME

PROG_CHARGE_OPTIONS = {
    0: "No Grid or Gen",
    1: "Allow Grid",
    2: "Allow Gen",
    3: "Allow Grid & Gen",
}
PROG_MODE_OPTIONS = {
    0: "None",
    4: "General",
    8: "Backup",
    16: "Charge",
}

SENSORS += (
    PROG1_TIME,
    PROG2_TIME,
    PROG3_TIME,
    PROG4_TIME,
    PROG5_TIME,
    PROG6_TIME,
    NumberRWSensor(256, "Prog1 power", WATT, max=RATED_POWER),
    NumberRWSensor(257, "Prog2 power", WATT, max=RATED_POWER),
    NumberRWSensor(258, "Prog3 power", WATT, max=RATED_POWER),
    NumberRWSensor(259, "Prog4 power", WATT, max=RATED_POWER),
    NumberRWSensor(260, "Prog5 power", WATT, max=RATED_POWER),
    NumberRWSensor(261, "Prog6 power", WATT, max=RATED_POWER),
    NumberRWSensor(268, "Prog1 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(269, "Prog2 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(270, "Prog3 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(271, "Prog4 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(272, "Prog5 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(273, "Prog6 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    SelectRWSensor(274, "Prog1 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(275, "Prog2 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(276, "Prog3 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(277, "Prog4 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(278, "Prog5 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(279, "Prog6 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    SelectRWSensor(274, "Prog1 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    SelectRWSensor(275, "Prog2 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    SelectRWSensor(276, "Prog3 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    SelectRWSensor(277, "Prog4 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    SelectRWSensor(278, "Prog5 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    SelectRWSensor(279, "Prog6 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
)
SENSORS += (
    NumberRWSensor(
        262,
        "Prog1 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        263,
        "Prog2 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        264,
        "Prog3 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        265,
        "Prog4 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        266,
        "Prog5 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        267,
        "Prog6 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
)

#############
# Inverter settings
#############
SENSORS += NumberRWSensor(230, "Grid Charge Battery current", AMPS, min=0, max=185)
SENSORS += NumberRWSensor(210, "Battery Max Charge current", AMPS, min=0, max=185)
SENSORS += NumberRWSensor(211, "Battery Max Discharge current", AMPS, min=0, max=185)

#############
# Deprecated
#############


def _deprecated() -> None:
    """Populate the deprecated sensors."""
    dep_map: dict[str, Sensor] = {
        "aux_power": Sensor(166, "AUX load", WATT, -1),
        "battery_temperature": TempSensor(182, "Temp Battery", CELSIUS, 0.1),
        "dc_transformer_temperature": TempSensor(
            90, "Temp DC transformer", CELSIUS, 0.1
        ),
        "environment_temperature": TempSensor(95, "Temp Environment", CELSIUS, 0.1),
        "grid_charge_battery_current": Sensor(230, "Battery Grid Charge", AMPS, -1),
        "grid_ct_power": Sensor(172, "Grid CT load", WATT, -1),
        "grid_l2_power": Sensor(168, "Grid L2 load", WATT, -1),
        "grid_ld_power": Sensor(167, "Grid L1 load", WATT, -1),
        "grid_power": Sensor(169, "Grid load", WATT, -1),
        "inverter_power": Sensor(175, "Inverter Output", WATT, -1),
        "radiator_temperature": TempSensor(91, "Temp Radiator", CELSIUS, 0.1),
        "day_active_energy": Sensor(60, "Day Active Power", KWH, -0.1),
        "day_reactive_energy": Sensor(61, "Day Reactive Power", "kVarh", -0.1),
        "total_active_energy": Sensor((63, 64), "Total Active Power", KWH, 0.1),
        "month_pv_energy": Sensor(65, "Month PV Power", KWH),
        "month_load_energy": Sensor(66, "Month Load Power", KWH),
        "month_grid_energy": Sensor(67, "Month Grid Power", KWH),
        "year_pv_energy": Sensor((68, 69), "Year PV Power", KWH, 0.1),
        "day_load_energy": Sensor(84, "Day Load Power", KWH, 0.1),
        "total_load_energy": Sensor((85, 86), "Total Load Power", KWH, 0.1),
        "year_load_energy": Sensor((87, 88), "Year Load Power", KWH, 0.1),
        "total_pv_energy": Sensor((96, 97), "Total PV Power", KWH, 0.1),
        "battery_equalization_voltage": Sensor(201, "Equalization voltage", VOLT, 0.01),
        "battery_absorption_voltage": Sensor(202, "Absorption voltage", VOLT, 0.01),
    }

    for newname, sen in dep_map.items():
        SENSORS.deprecated[sen.id] = SENSORS.all[newname]


_deprecated()



"""Sunsynk sensor tests."""
import logging
from typing import Sequence

import pytest

from sunsynk.definitions import AMPS, CELSIUS, SENSORS, VOLT, WATT
from sunsynk.rwsensors import RWSensor, SelectRWSensor
from sunsynk.sensors import (
    FaultSensor,
    InverterStateSensor,
    MathSensor,
    SDStatusSensor,
    Sensor,
    SerialSensor,
    TempSensor,
)
from sunsynk.state import group_sensors

_LOGGER = logging.getLogger(__name__)


def test_sen():
    s = Sensor(0, "S 1")
    a = [s]
    assert a[0] == s
    assert s.id == "s_1"

    # Test __hash__
    s2 = Sensor(0, "S 2")
    ss = {s, s}
    assert len(ss) == 1
    ss = {s, s2}
    assert len(ss) == 2

    # Test eq
    s22 = Sensor(0, "S 2")
    assert s != s2
    assert s22 == s2
    with pytest.raises(TypeError):
        assert s == 0


def test_sensor_hash():
    ss = {Sensor(0, "S 1"), Sensor(0, "S 1")}
    assert len(ss) == 1
    ss = {Sensor(0, "S 1"), Sensor(0, "S 2")}
    assert len(ss) == 2

    s = Sensor(0, "S 1")
    m = MathSensor((0, 1), "math1", factors=(1, 1))
    ss = {s, m}
    assert len(ss) == 2

    s = RWSensor((0, 1), "rwswitch")
    m = SelectRWSensor((0, 1), "switch")
    ss = {s, m}
    assert len(ss) == 2


def test_group() -> None:
    sen = [
        Sensor(10, "10"),
        Sensor(11, "11"),
        Sensor(12, "12"),
        Sensor(20, "20"),
    ]
    g = list(group_sensors(sen))
    assert g == [[10, 11, 12], [20]]

    assert len(list(group_sensors(None))) == 0


def test_group_max_size() -> None:
    sen = [
        Sensor(10, "10"),
        Sensor(11, "11"),
        Sensor(12, "12"),
        Sensor(13, "13"),
        Sensor(14, "14"),
        Sensor(15, "15"),
        Sensor(16, "16"),
    ]
    g = list(group_sensors(sen, max_group_size=3))
    assert g == [[10, 11, 12], [13, 14, 15], [16]]


def test_all_groups() -> None:
    s = [SENSORS.all[s] for s in SENSORS.all]
    for i in range(2, 6):
        _LOGGER.warning("waste with %d gap: %s", i, waste(group_sensors(s, i)))

    grp = group_sensors(s)

    grplen = [len(i) for i in grp]

    assert grplen[:1] == [6]
    # assert grplen[-1:] == [1]


def waste(groups) -> Sequence[int]:
    """Calculate amount of unused registers in this grouping."""
    return [sum(b - a for a, b in zip(g[:-1], g[1:])) for g in groups]


def test_ids() -> None:
    for name, sen in SENSORS.all.items():
        assert name == sen.id

        if sen.factor and sen.factor < 0 and len(sen.address) > 1:
            assert False, "only single signed supported"

        if sen.id in SENSORS.deprecated:
            continue
        assert sen.unit != AMPS or sen.name.endswith(" current")
        assert sen.unit != WATT or sen.name.endswith(" power")
        assert sen.unit != VOLT or sen.name.endswith(" voltage")
        assert sen.unit != CELSIUS or sen.name.endswith(" temperature")


def test_other_sensors() -> None:
    s = TempSensor(1, "", "", 0.1)
    assert s.reg_to_value((1000,)) == 0

    s = SDStatusSensor(1, "", "")
    assert s.reg_to_value((1000,)) == "fault"
    assert s.reg_to_value((1,)) == "unknown 1"

    s = InverterStateSensor(1, "", "")
    assert s.reg_to_value((2,)) == "ok"
    assert s.reg_to_value((1,)) == "unknown 1"


def test_math() -> None:
    s = MathSensor((1, 2), "", "", factors=(1, -1))
    assert s.reg_to_value((1000, 800)) == 200

    with pytest.raises(AssertionError):
        MathSensor((0, 1), "", "", factors=(1))

    with pytest.raises(AssertionError):
        MathSensor((0, 1), "", "", factors="aaa")

    with pytest.raises(TypeError):
        MathSensor((0, 1), "", "")

    assert s.reg_to_value((200, 800)) == -600
    s.no_negative = True
    assert s.reg_to_value((200, 800)) == 0
    s.no_negative = False
    s.absolute = True
    assert s.reg_to_value((200, 800)) == 600
    s = MathSensor((1, 2), "", "", factors=(1, -1), absolute=True)
    assert s.reg_to_value((200, 800)) == 600


def test_update_func() -> None:
    s = SerialSensor(1, "", "")
    regs = (0x4148, 0x3738)
    assert s.reg_to_value(regs) == "AH78"


# def test_update_float(caplog) -> None:
#     s = TempSensor(60, "two", factor=0.1)
#     rmap = register_map(60, [1001])
#     update_sensors([s], rmap)

#     assert s.value == 0.1

#     s.reg_value = (None,)
#     s.update_value()
#     assert "Could not decode" in caplog.text


def test_decode_fault() -> None:
    s = FaultSensor(1, "", "")
    regs = (0x01, 0x0, 0x0, 0x0)
    assert s.reg_to_value(regs) == "F01"
    regs = (0x02, 0x0, 0x0, 0x0)
    assert s.reg_to_value(regs) == "F02"
    regs = (0x80, 0x0, 0x0, 0x0)
    assert s.reg_to_value(regs) == "F08"

    regs = (0x0, 0x8000, 0x0, 0x0)
    assert s.reg_to_value(regs) == "F32"
    regs = (0x0, 0x0, 0x1, 0x0)
    assert s.reg_to_value(regs) == "F33"



"""Sunsynk 5kW&8kW hybrid 3-phase inverter sensor definitions."""
# pylint: disable=duplicate-code
from sunsynk import AMPS, CELSIUS, KWH, VOLT, WATT
from sunsynk.rwsensors import NumberRWSensor, SelectRWSensor, TimeRWSensor
from sunsynk.sensors import (
    FaultSensor,
    InverterStateSensor,
    MathSensor,
    SDStatusSensor,
    Sensor,
    SensorDefinitions,
    SerialSensor,
    TempSensor,
)

SENSORS = SensorDefinitions()

##########
# Battery
##########
SENSORS += (
    TempSensor(586, "Battery temperature", CELSIUS, 0.1),
    Sensor(587, "Battery voltage", VOLT, 0.01),
    Sensor(588, "Battery SOC", "%"),
    Sensor(590, "Battery power", WATT, -1),
    Sensor(591, "Battery current", AMPS, -0.01),
)

#################
# Inverter Power
#################
SENSORS += (
    Sensor(636, "Inverter power", WATT, -1),
    Sensor(633, "Inverter L1 power", WATT, -1),
    Sensor(634, "Inverter L2 power", WATT, -1),
    Sensor(635, "Inverter L3 power", WATT, -1),
    Sensor(627, "Inverter voltage", VOLT, 0.1),
    Sensor(638, "Inverter frequency", "Hz", 0.01),
)

#############
# Grid Power
#############
SENSORS += (
    Sensor(609, "Grid frequency", "Hz", 0.01),
    Sensor(625, "Grid power", WATT, -1),  # gridTotalPac
    Sensor(622, "Grid L1 power", WATT, -1),  # aPower
    Sensor(623, "Grid L2 power", WATT, -1),  # bPower
    Sensor(624, "Grid L3 power", WATT, -1),  # cPower
    Sensor(600, "Grid voltage", VOLT, 0.1),  # aLineVolt
    MathSensor(
        (610, 611, 612), "Grid current", AMPS, factors=(0.01, 0.01, 0.01)
    ),  # iac1,iac2,iac3
    Sensor(619, "Grid CT power", WATT, -1),  # gridOutsideTotalPac (unsure ??)
)

#############
# Load Power
#############
SENSORS += (
    Sensor(653, "Load power", WATT, -1),
    Sensor(650, "Load L1 power", WATT, -1),
    Sensor(651, "Load L2 power", WATT, -1),
    Sensor(652, "Load L3 power", WATT, -1),
)

################
# Solar Power 1
################
SENSORS += (
    Sensor(672, "PV1 power", WATT, -1),
    Sensor(676, "PV1 voltage", VOLT, 0.1),
    Sensor(677, "PV1 current", AMPS, 0.1),
)

################
# Solar Power 2
################
SENSORS += (
    Sensor(673, "PV2 power", WATT, -1),
    Sensor(678, "PV2 voltage", VOLT, 0.1),
    Sensor(679, "PV2 current", AMPS, 0.1),
)

################
# Solar Power 3 (?? deye 3 Phase  only has 2 MPPT)
################
SENSORS += (
    Sensor(674, "PV3 power", WATT, -1),
    Sensor(680, "PV3 voltage", VOLT, 0.1),
    Sensor(681, "PV3 current", AMPS, 0.1),
)

################
# Solar Power 4 (?? deye 3 Phase  only has 2 MPPT)
################
SENSORS += (
    Sensor(675, "PV4 power", WATT, -1),
    Sensor(682, "PV4 voltage", VOLT, 0.1),
    Sensor(683, "PV4 current", AMPS, 0.1),
)


###################
# Power on Outputs (aka GEN - which has multiple modes on the deye 3 Phase )
###################
SENSORS += (
    Sensor(667, "Gen power", WATT, -1),
    Sensor(664, "Gen L1 power", WATT, -1),
    Sensor(665, "Gen L2 power", WATT, -1),
    Sensor(666, "Gen L3 power", WATT, -1),
)

###################
# Energy
###################
SENSORS += (
    Sensor(502, "Day Active Energy", KWH, -0.1),
    Sensor(514, "Day Battery Charge", KWH, 0.1),
    Sensor(515, "Day Battery discharge", KWH, 0.1),
    Sensor(521, "Day Grid Export", KWH, 0.1),
    Sensor(520, "Day Grid Import", KWH, 0.1),
    Sensor(536, "Day Gen Energy", KWH, 0.1),
    Sensor(526, "Day Load Energy", KWH, 0.1),  # I guess "used" = load?
    Sensor(529, "Day PV Energy", KWH, 0.1),
    Sensor((506, 507), "Total Active Energy", KWH, 0.1),  # signed?
    Sensor((516, 517), "Total Battery Charge", KWH, 0.1),
    Sensor((518, 519), "Total Battery Discharge", KWH, 0.1),
    Sensor((524, 525), "Total Grid Export", KWH, 0.1),
    Sensor((522, 523), "Total Grid Import", KWH, 0.1),
    Sensor((527, 528), "Total Load Energy", KWH, 0.1),
    Sensor((534, 535), "Total PV Energy", KWH, 0.1),
)

##########
# General
##########
RATED_POWER = Sensor((20, 21), "Rated power", WATT, 0.1)
SENSORS += RATED_POWER

SENSORS += (
    Sensor(
        0, "Device Type"
    ),  # {2: "Inverter", 3: "Hybrid Inverter", 4: "Micro Inverter", 5: "3 Phase Hybrid Inverter" }),
    FaultSensor((555, 556, 557, 558), "Fault"),
    InverterStateSensor(500, "Overall state"),
    SDStatusSensor(0, "SD Status", ""),  # type: ignore        # 3 Phase does not have SD Card but crashes when removed
    SerialSensor((3, 4, 5, 6, 7), "Serial"),
    TempSensor(540, "DC transformer temperature", CELSIUS, 0.1),
    TempSensor(541, "Radiator temperature", CELSIUS, 0.1),
    Sensor(
        552, "Grid Connected Status"
    ),  # Bit 0 = INV Relay, 1 = Undef, 2 = Grid relay, 3 = Gen relay, 4 = Grid give, 5, Dry contact
)


###########
# Settings
###########
SENSORS += (
    NumberRWSensor(128, "Grid Charge Battery current", AMPS, max=210),
    SelectRWSensor(130, "Grid Charge enabled", switch=(0, 1)),
    SelectRWSensor(146, "Use Timer", switch=(0, 255)),
    SelectRWSensor(145, "Solar Export", switch=(0, 1)),
    NumberRWSensor(143, "Export Limit power", WATT, max=RATED_POWER),
    NumberRWSensor(108, "Battery Max Charge current", AMPS, max=210),
    NumberRWSensor(109, "Battery Max Discharge current", AMPS, max=210),
    NumberRWSensor(102, "Battery Capacity current", AMPS, max=2000),
)

# Additional optional sensors
SENSORS += (
    NumberRWSensor(
        103, "Battery low voltage", VOLT, -0.1
    ),  # interesting perhaps not available in menu. default 4500 (45V)
    SelectRWSensor(
        133,
        "Generator Port Usage",
        options={0: "Generator", 1: "Smartload", 2: "Micro Inverter"},
    ),
    SelectRWSensor(129, "Generator Charge enabled", options={0: "Off", 1: "On"}),
    SelectRWSensor(112, "Battery Activate", options={0: "Off", 1: "On"}),
    NumberRWSensor(113, "Battery Resistance", "mΩ", max=6000),
    NumberRWSensor(104, "System Zero Export power", WATT, -1),
    NumberRWSensor(105, "Battery Equalization Days", "days", -1),
    NumberRWSensor(106, "Battery Equalization Hours", "h", -1),  # 1 = 0.5 hours
)

# Absolute min and max voltage based on Deye inverter
MIN_VOLTAGE = 41
MAX_VOLTAGE = 60

BATTERY_EQUALIZATION_VOLTAGE = NumberRWSensor(
    99, "Battery Equalization voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)
BATTERY_ABSORPTION_VOLTAGE = NumberRWSensor(
    100, "Battery Absorption voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)
BATTERY_FLOAT_VOLTAGE = NumberRWSensor(
    101, "Battery Float voltage", VOLT, 0.01, min=MIN_VOLTAGE, max=MAX_VOLTAGE
)

BATTERY_SHUTDOWN_CAPACITY = NumberRWSensor(115, "Battery Shutdown Capacity", "%")
BATTERY_RESTART_CAPACITY = NumberRWSensor(116, "Battery Restart Capacity", "%")
BATTERY_LOW_CAPACITY = NumberRWSensor(
    117,
    "Battery Low Capacity",
    "%",
    min=BATTERY_SHUTDOWN_CAPACITY,
    max=BATTERY_RESTART_CAPACITY,
)
BATTERY_SHUTDOWN_CAPACITY.max = BATTERY_LOW_CAPACITY
BATTERY_RESTART_CAPACITY.min = BATTERY_LOW_CAPACITY

BATTERY_SHUTDOWN_VOLTAGE = NumberRWSensor(
    118, "Battery Shutdown voltage", VOLT, 0.01, min=MIN_VOLTAGE
)
BATTERY_RESTART_VOLTAGE = NumberRWSensor(
    119, "Battery Restart voltage", VOLT, 0.01, max=MAX_VOLTAGE
)
BATTERY_LOW_VOLTAGE = NumberRWSensor(
    120,
    "Battery Low voltage",
    VOLT,
    0.01,
    min=BATTERY_SHUTDOWN_VOLTAGE,
    max=BATTERY_RESTART_VOLTAGE,
)
BATTERY_SHUTDOWN_VOLTAGE.max = BATTERY_LOW_VOLTAGE
BATTERY_RESTART_VOLTAGE.min = BATTERY_LOW_VOLTAGE

SENSORS += (
    BATTERY_EQUALIZATION_VOLTAGE,
    BATTERY_ABSORPTION_VOLTAGE,
    BATTERY_FLOAT_VOLTAGE,
    BATTERY_SHUTDOWN_CAPACITY,
    BATTERY_RESTART_CAPACITY,
    BATTERY_LOW_CAPACITY,
    BATTERY_SHUTDOWN_VOLTAGE,
    BATTERY_RESTART_VOLTAGE,
    BATTERY_LOW_VOLTAGE,
)

#################
# System program
#################

SENSORS += SelectRWSensor(
    141, "Priority Mode", options={0: "Battery first", 1: "Load first"}
)

SENSORS += SelectRWSensor(
    142,
    "Load Limit",
    options={0: "Allow Export", 1: "Essentials", 2: "Zero Export"},
)

PROG1_TIME = TimeRWSensor(148, "Prog1 Time")
PROG2_TIME = TimeRWSensor(149, "Prog2 Time", min=PROG1_TIME)
PROG3_TIME = TimeRWSensor(150, "Prog3 Time", min=PROG2_TIME)
PROG4_TIME = TimeRWSensor(151, "Prog4 Time", min=PROG3_TIME)
PROG5_TIME = TimeRWSensor(152, "Prog5 Time", min=PROG4_TIME)
PROG6_TIME = TimeRWSensor(153, "Prog6 Time", min=PROG5_TIME)
PROG1_TIME.min = PROG6_TIME
PROG1_TIME.max = PROG2_TIME
PROG2_TIME.max = PROG3_TIME
PROG3_TIME.max = PROG4_TIME
PROG4_TIME.max = PROG5_TIME
PROG5_TIME.max = PROG6_TIME
PROG6_TIME.max = PROG1_TIME

PROG_CHARGE_OPTIONS = {
    0: "Off",
    1: "On",
}

# PROG_CHARGE_OPTIONS = {
#     0: "No Grid or Gen",
#     1: "Allow Grid",
#     2: "Allow Gen",
#     3: "Allow Grid & Gen",
# }
# PROG_MODE_OPTIONS = {
#     0: "None",
#     4: "General",
#     8: "Backup",
#     16: "Charge",
# }

SENSORS += (
    PROG1_TIME,
    PROG2_TIME,
    PROG3_TIME,
    PROG4_TIME,
    PROG5_TIME,
    PROG6_TIME,
    NumberRWSensor(154, "Prog1 power", WATT, max=RATED_POWER),
    NumberRWSensor(155, "Prog2 power", WATT, max=RATED_POWER),
    NumberRWSensor(156, "Prog3 power", WATT, max=RATED_POWER),
    NumberRWSensor(157, "Prog4 power", WATT, max=RATED_POWER),
    NumberRWSensor(158, "Prog5 power", WATT, max=RATED_POWER),
    NumberRWSensor(159, "Prog6 power", WATT, max=RATED_POWER),
    NumberRWSensor(166, "Prog1 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(167, "Prog2 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(168, "Prog3 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(169, "Prog4 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(170, "Prog5 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    NumberRWSensor(171, "Prog6 Capacity", "%", min=BATTERY_LOW_CAPACITY),
    SelectRWSensor(172, "Prog1 charge", options=PROG_CHARGE_OPTIONS),
    SelectRWSensor(173, "Prog2 charge", options=PROG_CHARGE_OPTIONS),
    SelectRWSensor(174, "Prog3 charge", options=PROG_CHARGE_OPTIONS),
    SelectRWSensor(175, "Prog4 charge", options=PROG_CHARGE_OPTIONS),
    SelectRWSensor(176, "Prog5 charge", options=PROG_CHARGE_OPTIONS),
    SelectRWSensor(177, "Prog6 charge", options=PROG_CHARGE_OPTIONS),
    # SelectRWSensor(172, "Prog1 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(173, "Prog2 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(174, "Prog3 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(175, "Prog4 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(176, "Prog5 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(177, "Prog6 charge", options=PROG_CHARGE_OPTIONS, bitmask=0x03),
    # SelectRWSensor(172, "Prog1 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    # SelectRWSensor(173, "Prog2 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    # SelectRWSensor(174, "Prog3 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    # SelectRWSensor(175, "Prog4 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    # SelectRWSensor(176, "Prog5 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
    # SelectRWSensor(177, "Prog6 mode", options=PROG_MODE_OPTIONS, bitmask=0x1C),
)


SENSORS += (
    NumberRWSensor(
        160,
        "Prog1 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        161,
        "Prog2 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        162,
        "Prog3 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        163,
        "Prog4 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        164,
        "Prog5 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
    NumberRWSensor(
        165,
        "Prog6 voltage",
        VOLT,
        0.01,
        min=BATTERY_LOW_VOLTAGE,
        max=BATTERY_FLOAT_VOLTAGE,
    ),
)



from unittest.mock import call

from sunsynk.rwsensors import NumberRWSensor, Sensor
from sunsynk.state import InverterState


def test_on_changed(state: InverterState) -> None:
    """Test update and on_change."""
    onchange = state.onchange

    s1 = NumberRWSensor((1,), "S1", min=1, max=50)
    state.track(s1)

    assert state.values[s1] is None

    state.update({1: 5})
    assert state.values[s1] == 5
    assert onchange.call_count == 1
    assert onchange.call_args == call(s1, 5, None)

    state.update({1: 10})
    assert state.values[s1] == 10
    assert onchange.call_count == 2
    assert onchange.call_args == call(s1, 10, 5)

    s2 = Sensor(1, "S2", "")
    state.track(s2)
    assert state.values[s2] is None

    onchange.call_args_list.clear()
    state.update({1: 20})
    assert state.values[s1] == 20
    assert state.values[s2] == 20
    assert onchange.call_count == 4
    assert onchange.call_args_list == [
        call(s1, 20, 10),
        call(s2, 20, None),
    ]



"""Sunsynk lib using umodbus."""
import asyncio
import logging
from typing import Sequence
from urllib import parse

import attr
from async_modbus import AsyncClient, modbus_for_url
from connio import SERIAL_SCHEMES, SOCKET_SCHEMES

from sunsynk.sunsynk import Sunsynk

_LOGGER = logging.getLogger(__name__)


@attr.define
class uSunsynk(Sunsynk):  # pylint: disable=invalid-name
    """Sunsynk class using umodbus."""

    client: AsyncClient = attr.field(default=None)

    async def connect(self) -> None:
        """Connect."""
        url_result = parse.urlparse(self.port)
        scheme = url_result.scheme
        if scheme not in SOCKET_SCHEMES and scheme not in SERIAL_SCHEMES:
            sks = [f"{s}://" for s in SOCKET_SCHEMES]
            sks.extend(f"{s}://" for s in SERIAL_SCHEMES)
            raise ValueError(
                "Bad port/connection URL. It should start with one of the following: "
                + ", ".join(sks)
            )

        conn_opt = {
            "timeout": self.timeout,  # sockio.TCP & Serial._timeout
        }
        if scheme and scheme in SOCKET_SCHEMES:
            conn_opt["connection_timeout"] = self.timeout  # sockio.TCP

        self.client = modbus_for_url(self.port, conn_opt)

    async def write_register(self, *, address: int, value: int) -> bool:
        """Write to a register - Sunsynk supports modbus function 0x10."""
        try:
            await asyncio.wait_for(
                self.client.write_registers(
                    slave_id=self.server_id,
                    starting_address=address,
                    values=(value,),
                ),
                timeout=self.timeout,
            )
            return True
        except asyncio.TimeoutError:
            _LOGGER.error("timeout writing register %s=%s", address, value)
        self.timeouts += 1
        return False

    async def read_holding_registers(self, start: int, length: int) -> Sequence[int]:
        """Read a holding register."""
        return await self.client.read_holding_registers(self.server_id, start, length)



"""Sunsynk sensor tests."""
import logging

import pytest

from sunsynk.rwsensors import (
    NumberRWSensor,
    RWSensor,
    SelectRWSensor,
    Sensor,
    TimeRWSensor,
)

_LOGGER = logging.getLogger(__name__)


def test_bitmask(caplog, state) -> None:
    s = RWSensor(1, "", bitmask=0x1)
    with pytest.raises(NotImplementedError):
        s.value_to_reg(None, state.get)

    s = NumberRWSensor(1, "", min=1, max=10, bitmask=0x1)

    state.track(s)
    assert state[s] is None

    state.update({1: 1})
    assert state[s] == 1
    assert "outside" not in caplog.text

    state.update({1: 3})

    val = 3
    reg = s.value_to_reg(val, state.get)
    reg = s.reg(*reg, msg=f"value was {val}")

    assert reg == (1,)
    assert "outside" in caplog.text


def test_bitmask2(caplog, state) -> None:
    s = SelectRWSensor(1, "", switch=(0, 4), bitmask=0x4)
    state.track(s)

    assert state[s] is None
    state.update({1: 0x1})

    assert state[s] == "OFF"
    assert "outside" not in caplog.text

    state.update({1: 0x14})
    assert state[s] == "ON"

    val = "OFF"
    reg = s.value_to_reg(val, state.get)
    reg = s.reg(*reg, msg=f"value was {val}")
    assert reg == (0,)


def test_number_rw(state) -> None:
    s = NumberRWSensor(1, "s1", min=1, max=10, factor=1)

    assert s.dependencies == []
    assert s.min == 1
    assert s.max == 10

    s.min = Sensor(2, "2")
    assert s.dependencies == [s.min]

    s.max = Sensor(3, "3")
    assert s.dependencies == [s.min, s.max]

    state.track(s)

    state.update(
        {
            1: 230,
            2: 1,
            3: 300,
        }
    )
    assert state[s] == 230

    state.update({1: 500})
    assert state[s] == 500  # does not take min/max into account!

    assert s.value_to_reg(200, state.get) == (200,)
    assert s.value_to_reg(500, state.get) == (300,)
    assert s.value_to_reg(-1, state.get) == (1,)

    s = NumberRWSensor(1, "s2", factor=0.01)
    state.track(s)

    state.update({1: 4850})
    assert state.registers[1] == 4850
    assert state[s] == 48.5

    long = NumberRWSensor((1, 2), "long", max=0xFFFFF)
    reg55 = (55, 1)
    res55 = (1 << 16) + 55
    assert long.reg_to_value(reg55) == res55
    assert long.value_to_reg(res55, state.get) == reg55


def test_select_rw(caplog, state) -> None:
    s = SelectRWSensor(1, "", options={1: "one", 2: "two"})

    assert s.reg_to_value((2,)) == "two"
    assert s.value_to_reg("two", state.get) == (2,)
    assert s.reg_to_value((5,)) is None

    state.track(s)
    state.update({1: 1})
    assert state[s] == "one"

    state.update({1: 2})
    assert state[s] == "two"

    assert s.available_values() == ["one", "two"]

    assert s.value_to_reg("five", state.get) == (2,)
    assert caplog.records[-1].message == "Unknown five"
    assert caplog.records[-1].levelname == "WARNING"


def test_time_rw(state) -> None:
    s = TimeRWSensor(60, "two", factor=0.1)
    state.track(s)

    state.update(
        {
            60: 300,
        }
    )
    assert state[s] == "3:00"

    assert s.value_to_reg("0:00", state.get) == (0,)
    assert s.value_to_reg("4:01", state.get) == (401,)
    assert s.value_to_reg("23:59", state.get) == (2359,)

    assert s.dependencies == []
    s.min = TimeRWSensor(50, "min", factor=0.1)
    assert s.dependencies == [s.min]
    s.max = TimeRWSensor(70, "max", factor=0.1)
    assert s.dependencies == [s.min, s.max]

    state.track(s.min, s.max)
    assert state[s.min] is None
    assert state[s.max] is None
    state.update({50: 200, 70: 300})
    assert state[s.min] == "2:00"
    assert state[s.max] == "3:00"

    assert s.available_values(15, state.get) == ["2:00", "2:15", "2:30", "2:45", "3:00"]

    s.reg_to_value((201,))
    assert s.available_values(15, state.get) == [
        "2:00",
        "2:15",
        "2:30",
        "2:45",
        "3:00",
    ]

    state.update(
        {
            50: 2330,
            60: 2330,
            70: 30,
        }
    )
    assert s.available_values(15, state.get) == [
        "23:30",
        "23:45",
        "0:00",
        "0:15",
        "0:30",
    ]

    state.update(
        {
            50: 200,
            70: 200,
        }
    )
    assert s.available_values(15, state.get) == ["2:00"]


# def test_update_sensor(caplog) -> None:
#     s = NumberRWSensor(60, "two", factor=0.1)
#     assert s.value is None
#     update_sensors([s], {})
#     assert s.value is None
#     update_sensors([s], {60: 10})
#     assert s.value == 1


def test_bad_sensor(caplog) -> None:
    NumberRWSensor((60, 1), "two", factor=0.1, bitmask=1)
    assert "single register" in caplog.text
